from __future__ import annotations
from typing import Tuple, Optional, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import numpy.typing as npt

from ._ies import InversionType


def _validate_inputs(
    response_ensemble: npt.NDArray[np.double],
    noise: Optional[npt.NDArray[np.double]],
    observation_errors: npt.NDArray[np.double],
    observation_values: npt.NDArray[np.double],
    param_ensemble: Optional[npt.NDArray[np.double]] = None,
) -> None:
    if len(response_ensemble.shape) != 2:
        raise ValueError(
            "response_ensemble must be a matrix of size (number of responses by number of realizations)"
        )

    num_responses = response_ensemble.shape[0]
    ensemble_size = response_ensemble.shape[1]

    if response_ensemble.shape[1] != ensemble_size:
        raise ValueError(
            "response_ensemble and parameter_ensemble must have the same number of columns"
        )

    if noise is not None and noise.shape[1] != ensemble_size:
        raise ValueError(
            "noise and response_ensemble must have the same number of columns"
        )

    if noise is not None and noise.shape[0] != num_responses:
        raise ValueError(
            "noise and response_ensemble must have the same number of rows"
        )

    if len(observation_errors) != len(observation_values):
        raise ValueError(
            "observation_errors and observation_values must have the same number of elements"
        )

    if len(observation_values) != num_responses:
        raise ValueError(
            "observation_values must have the same number of elements as there are responses"
        )

    if param_ensemble is not None and len(param_ensemble.shape) != 2:
        raise ValueError(
            "parameter_ensemble must be a matrix of size (number of parameters by number of realizations)"
        )

    if param_ensemble is not None and param_ensemble.shape[1] != ensemble_size:
        raise ValueError(
            "param_ensemble and response_ensemble must have the same number of columns"
        )


def _create_errors(
    observation_errors: npt.NDArray[np.double],
    inversion: InversionType,
) -> Tuple[npt.NDArray[np.double], npt.NDArray[np.double]]:
    if len(observation_errors.shape) == 2:
        R = observation_errors
        observation_errors = np.sqrt(observation_errors.diagonal())
        R = (R.T / R.diagonal()).T
    elif len(observation_errors.shape) == 1 and inversion == InversionType.EXACT_R:
        R = np.identity(len(observation_errors))
    else:
        R = None
    return R, observation_errors



import numpy as np
import pytest
import pandas as pd
from p_tqdm import p_map

from iterative_ensemble_smoother._ies import make_D
import iterative_ensemble_smoother as ies

rng = np.random.default_rng()

# The following tests follow the
# posterior properties described in
# https://ert.readthedocs.io/en/latest/theory/ensemble_based_methods.html#kalman-posterior-properties
a_true = 1.0
b_true = 5.0
number_of_parameters = 2
number_of_observations = 45


class LinearModel:
    def __init__(self, a, b):
        self.a = a
        self.b = b
        self.size = 2

    @classmethod
    def random(cls):
        a_std = 2.0
        b_std = 2.0
        # Priors with bias
        a_bias = 0.5 * a_std
        b_bias = -0.5 * b_std

        return cls(
            rng.normal(a_true + a_bias, a_std),
            rng.normal(b_true + b_bias, b_std),
        )

    def eval(self, x):
        return self.a * x + self.b


@pytest.mark.parametrize("number_of_realizations", [100, 200])
def test_that_es_update_for_a_linear_model_follows_theory(number_of_realizations):
    true_model = LinearModel(a_true, b_true)

    ensemble = [LinearModel.random() for _ in range(number_of_realizations)]

    A = np.array(
        [
            [realization.a for realization in ensemble],
            [realization.b for realization in ensemble],
        ]
    )
    mean_prior = np.mean(A, axis=1)

    # We use time as the x-axis and observations are at
    # t=0,1,2...number_of_observations
    times = np.arange(number_of_observations)

    S = np.array([[realization.eval(t) for realization in ensemble] for t in times])

    # When observations != true model, then ml estimates != true parameters.
    # This gives both a more advanced and realistic test. Standard normal
    # N(0,1) noise is added to obtain this. The randomness ensures we are not
    # gaming the test. But the difference could in principle be any non-zero
    # scalar.
    observations = np.array(
        [true_model.eval(t) + np.random.normal(0.0, 1.0) for t in times]
    )

    # Leading to fixed Maximum likelihood estimate.
    # It will equal true values when observations are sampled without noise.
    # It will also stay the same over beliefs.
    mean_observations = np.mean(observations)
    times_mean = np.mean(times)
    times_square_sum = sum(np.square(times))
    a_maximum_likelihood = sum(
        t * (observations[t] - mean_observations) for t in times
    ) / (times_square_sum - times_mean * sum(times))
    b_maximum_likelihood = mean_observations - a_maximum_likelihood * times_mean
    maximum_likelihood = np.array([a_maximum_likelihood, b_maximum_likelihood])

    previous_mean_posterior = mean_prior

    # numerical precision tolerance
    epsilon = 1e-2

    # We iterate with an increased belief in the observations
    for error in [10000.0, 100.0, 10.0, 1.0, 0.1]:
        # An important point here is that we do not iteratively
        # update A, but instead, observations stay the same and
        # we increase our belief in the observations
        # As A is update inplace, we have to reset it.
        A = np.array(
            [
                [realization.a for realization in ensemble],
                [realization.b for realization in ensemble],
            ]
        )
        smoother = ies.ES()
        smoother.fit(
            S,
            np.full(observations.shape, error),
            observations,
        )
        A_posterior = smoother.update(A)
        mean_posterior = np.mean(A_posterior, axis=1)

        # All posterior estimates lie between prior and maximum likelihood estimate
        assert np.all(
            np.linalg.norm(mean_posterior - maximum_likelihood)
            - np.linalg.norm(mean_prior - maximum_likelihood)
            < epsilon
        )
        assert np.all(
            np.linalg.norm(mean_prior - mean_posterior)
            - np.linalg.norm(mean_prior - maximum_likelihood)
            < epsilon
        )

        # Posterior parameter estimates improve with increased trust in observations
        assert np.all(
            np.linalg.norm(mean_posterior - maximum_likelihood)
            - np.linalg.norm(previous_mean_posterior - maximum_likelihood)
            < epsilon
        )

        previous_mean_posterior = mean_posterior

    # At strong beliefs, we should be close to the maximum likelihood estimate
    assert np.all(
        np.linalg.norm(previous_mean_posterior - maximum_likelihood) < epsilon
    )


def test_that_sies_converges_to_es_in_gauss_linear_case():
    def poly(a, b, c, x):
        return a * x**2 + b * x + c

    ensemble_size = 200

    # True patameter values
    a_t = 0.5
    b_t = 1.0
    c_t = 3.0

    noise_scale = 0.1
    x_observations = [0, 2, 4, 6, 8]
    observations = [
        (
            poly(a_t, b_t, c_t, x)
            + rng.normal(loc=0, scale=noise_scale * poly(a_t, b_t, c_t, x)),
            noise_scale * poly(a_t, b_t, c_t, x),
            x,
        )
        for x in x_observations
    ]

    d = pd.DataFrame(observations, columns=["value", "sd", "x"])
    d = d.set_index("x")
    num_obs = d.shape[0]

    # Assume diagonal ensemble covariance matrix for the measurement perturbations.
    R = np.diag(d.sd.values**2)

    E = rng.multivariate_normal(mean=np.zeros(len(R)), cov=R, size=ensemble_size).T
    assert E.shape == (num_obs, ensemble_size)

    coeff_a = rng.normal(0, 1, size=ensemble_size)
    coeff_b = rng.normal(0, 1, size=ensemble_size)
    coeff_c = rng.normal(0, 1, size=ensemble_size)

    X = np.concatenate(
        (coeff_a.reshape(-1, 1), coeff_b.reshape(-1, 1), coeff_c.reshape(-1, 1)), axis=1
    ).T

    fwd_runs = p_map(
        poly,
        coeff_a,
        coeff_b,
        coeff_c,
        [np.arange(max(x_observations) + 1)] * ensemble_size,
        desc="Running forward model.",
    )

    # Pick responses where we have observations
    response_ensemble = np.array(
        [fwd_run[d.index.get_level_values("x").to_list()] for fwd_run in fwd_runs]
    ).T

    smoother_es = ies.ES()
    smoother_es.fit(response_ensemble, d.sd.values, d.value.values, noise=E)
    params_es = smoother_es.update(X)

    assert np.linalg.det(np.cov(X)) > np.linalg.det(np.cov(params_es))

    params_ies = X.copy()
    responses_ies = response_ensemble.copy()
    smoother_ies = ies.SIES(ensemble_size)
    for _ in range(10):
        smoother_ies.fit(responses_ies, d.sd.values, d.value.values, noise=E)
        params_ies = smoother_ies.update(params_ies)

        _coeff_a = params_ies[0, :]
        _coeff_b = params_ies[1, :]
        _coeff_c = params_ies[2, :]

        _fwd_runs = p_map(
            poly,
            _coeff_a,
            _coeff_b,
            _coeff_c,
            [np.arange(max(x_observations) + 1)] * ensemble_size,
            desc="SIES ert iteration",
        )

        assert np.linalg.det(np.cov(X)) > np.linalg.det(np.cov(params_ies))

        responses_ies = np.array(
            [fwd_run[d.index.get_level_values("x").to_list()] for fwd_run in _fwd_runs]
        ).T

    assert np.abs(
        (np.linalg.det(np.cov(params_es)) - np.linalg.det(np.cov(params_ies))) < 0.001
    )


var = 2.0


@pytest.mark.parametrize(
    "inversion,errors",
    [
        (ies.InversionType.EXACT, np.diag(np.array([var, var, var]))),
        (ies.InversionType.EXACT, np.array([np.sqrt(var), np.sqrt(var), np.sqrt(var)])),
        (ies.InversionType.EXACT_R, np.diag(np.array([var, var, var]))),
        (
            ies.InversionType.EXACT_R,
            np.array([np.sqrt(var), np.sqrt(var), np.sqrt(var)]),
        ),
        (ies.InversionType.SUBSPACE_RE, np.diag(np.array([var, var, var]))),
        (
            ies.InversionType.SUBSPACE_RE,
            np.array([np.sqrt(var), np.sqrt(var), np.sqrt(var)]),
        ),
    ],
)
def test_that_update_correctly_multiples_gaussians(inversion, errors):
    """
    NB! This test is potentially flaky because of finite ensemble size.

    Bayes' theorem states that
    p(x|y) is proportional to p(y|x)p(x)

    Assume p(x) is N(mu=0, Sigma=2I) and p(y|x) is N(mu=y, Sigma=2I).
    Multiplying these together (see 8.1.8 of the matrix cookbook) we get
    that p(x|y) is N(mu=y/2, Sigma=I).
    Note that Sigma is a covariance matrix.

    Here we use this property, and assume that the forward model is the identity
    to test analysis steps.
    """
    N = 1500
    nparam = 3

    A = rng.standard_normal(size=(nparam, N))
    A = np.linalg.cholesky(var * np.identity(nparam)) @ A
    # Assuming forward model is the identity
    Y = A

    obs_val = 10
    observation_values = np.array([obs_val, obs_val, obs_val])
    smoother = ies.SIES(N)
    smoother.fit(
        Y,
        errors,
        observation_values,
        inversion=inversion,
        step_length=1.0,
        truncation=1.0,
    )
    A_ES = smoother.update(A)

    for i in range(nparam):
        assert np.isclose(A_ES[i, :].mean(), obs_val / 2, rtol=0.15)

    assert (np.abs(np.cov(A_ES) - np.identity(nparam)) < 0.15).all()


def test_make_D():
    S = np.array([[2.0, 4.0], [6.0, 8.0]])
    E = np.array([[1.0, 2.0], [3.0, 4.0]])
    observation_values = np.array([1.0, 1.0])
    assert make_D(observation_values, E, S).tolist() == [
        [1.0 - 2 + 1.0, 2.0 - 4.0 + 1.0],
        [3.0 - 6.0 + 1.0, 4.0 - 8 + 1.0],
    ]


@pytest.mark.parametrize(
    "ensemble_size,num_params,linear",
    [
        pytest.param(100, 3, True, id="No projection because linear model"),
        pytest.param(
            100,
            3,
            False,
            id="Project because non-linear and num_params < ensemble_size - 1",
        ),
        pytest.param(3, 100, True, id="No projection because linear model"),
        pytest.param(
            3, 100, False, id="No projection because num_params > ensemble_size - 1"
        ),
        pytest.param(
            10, 10, True, id="No projection because num_params > ensemble_size - 1"
        ),
    ],
)
def test_that_global_es_update_is_identical_to_local(ensemble_size, num_params, linear):
    num_obs = num_params
    X = rng.normal(size=(num_params, ensemble_size))

    Y = X if linear else np.power(X, 2)

    observation_errors = rng.uniform(size=num_obs)
    observation_values = rng.normal(np.zeros(num_params), observation_errors)
    noise = rng.normal(size=(num_obs, ensemble_size))

    param_ensemble = None
    if not linear and num_params < ensemble_size - 1:
        param_ensemble = X

    smoother = ies.ES()
    smoother.fit(
        Y,
        observation_errors,
        observation_values,
        noise=noise,
        param_ensemble=param_ensemble,
    )
    X_ES_global = smoother.update(X)

    X_ES_local = np.zeros(shape=(num_params, ensemble_size))

    # First update a subset or batch of parameters at once
    # and and then loop through each remaining parameter and
    # update it separately.
    # This is meant to emulate a configuration where users
    # want to update, say, a field parameter separately from scalar parameters.
    batch = list(range(num_params // 3))
    smoother.fit(
        Y,
        observation_errors,
        observation_values,
        noise=noise,
        param_ensemble=param_ensemble,
    )
    X_ES_local[batch, :] = smoother.update(X[batch])

    for i in range(num_params // 3, num_params):
        smoother.fit(
            Y,
            observation_errors,
            observation_values,
            noise=noise,
            param_ensemble=param_ensemble,
        )
        X_ES_local[i, :] = smoother.update(X[i, :])

    assert np.isclose(X_ES_global, X_ES_local).all()


def test_that_ies_runs_with_failed_realizations():
    rng = np.random.default_rng()
    ensemble_size = 50
    num_params = 100
    num_responses = 5
    param_ensemble = rng.normal(size=(num_params, ensemble_size))
    response_ensemble = np.power(param_ensemble, 2)[:num_responses, :] + rng.normal(
        size=(num_responses, ensemble_size)
    )
    noise = rng.normal(size=(num_responses, ensemble_size))
    obs_values = rng.normal(size=num_responses)
    obs_errors = rng.normal(size=num_responses)
    ens_mask = np.array([True] * ensemble_size)
    ens_mask[10:] = False
    smoother = ies.SIES(ensemble_size)
    smoother.fit(
        response_ensemble[:, ens_mask],
        obs_errors,
        obs_values,
        noise=noise[:, ens_mask],
        ensemble_mask=ens_mask,
    )
    param_ensemble = smoother.update(param_ensemble[:, ens_mask])

    smoother.fit(
        response_ensemble[:, ens_mask],
        obs_errors,
        obs_values,
        noise=noise[:, ens_mask],
        ensemble_mask=ens_mask,
    )
    param_ensemble = smoother.update(param_ensemble)

    smoother.fit(
        response_ensemble[:, ens_mask],
        obs_errors,
        obs_values,
        noise=noise[:, ens_mask],
        ensemble_mask=ens_mask,
    )
    param_ensemble = smoother.update(param_ensemble)

    assert param_ensemble.shape == (num_params, ens_mask.sum())



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Fitting a polynomial with Gaussian priors
#
# We fit a simple polynomial with Gaussian priors, which is an example of a Gauss-linear problem for which the results obtained using Subspace Iterative Ensemble Smoother (SIES) tend to those obtained using Ensemble Smoother (ES).
# This notebook illustrated this property.
# %%
import numpy as np
import pandas as pd

np.set_printoptions(suppress=True)
rng = np.random.default_rng(12345)

import matplotlib.pyplot as plt

plt.rcParams["figure.figsize"] = (6, 6)
plt.rcParams.update({"font.size": 10})
from ipywidgets import interact
import ipywidgets as widgets

from p_tqdm import p_map

import iterative_ensemble_smoother as ies

# %% [markdown]
# ## Define synthetic truth and use it to create noisy observations

# %%
ensemble_size = 200


def poly(a, b, c, x):
    return a * x**2 + b * x + c


# True patameter values
a_t = 0.5
b_t = 1.0
c_t = 3.0

noise_scale = 0.01
x_observations = [0, 2, 4, 6, 8]
observations = [
    (
        poly(a_t, b_t, c_t, x)
        + rng.normal(loc=0, scale=noise_scale * poly(a_t, b_t, c_t, x)),
        noise_scale * poly(a_t, b_t, c_t, x),
        x,
    )
    for x in x_observations
]

d = pd.DataFrame(observations, columns=["value", "sd", "x"])
d = d.set_index("x")
num_obs = d.shape[0]

fig, ax = plt.subplots()
x_plot = np.linspace(0, 10, 50)
ax.set_title("Truth and noisy observations")
ax.set_xlabel("Time step")
ax.set_ylabel("Response")
ax.plot(x_plot, poly(a_t, b_t, c_t, x_plot))
ax.plot(d.index.get_level_values("x"), d.value.values, "o")
ax.grid()

# %% [markdown]
# ## Assume diagonal observation error covariance matrix and define perturbed observations

# %%
R = np.diag(d.sd.values**2)

E = rng.multivariate_normal(mean=np.zeros(len(R)), cov=R, size=ensemble_size).T
assert E.shape == (num_obs, ensemble_size)

D = np.ones((num_obs, ensemble_size)) * d.value.values.reshape(-1, 1) + E

# %% [markdown]
# ## Define Gaussian priors

# %%
coeff_a = rng.normal(0, 1, size=ensemble_size)
coeff_b = rng.normal(0, 1, size=ensemble_size)
coeff_c = rng.normal(0, 1, size=ensemble_size)

X = np.concatenate(
    (coeff_a.reshape(-1, 1), coeff_b.reshape(-1, 1), coeff_c.reshape(-1, 1)), axis=1
).T

# %% [markdown]
# ## Run forward model in parallel

# %%
fwd_runs = p_map(
    poly,
    coeff_a,
    coeff_b,
    coeff_c,
    [np.arange(max(x_observations) + 1)] * ensemble_size,
    desc=f"Running forward model.",
)

# %% [markdown]
# ## Pick responses where we have observations

# %%
Y = np.array(
    [fwd_run[d.index.get_level_values("x").to_list()] for fwd_run in fwd_runs]
).T

assert Y.shape == (
    num_obs,
    ensemble_size,
), "Measured responses must be a matrix with dimensions (number of observations x number of realisations)"

# %% [markdown]
# ## Condition on observations to calculate posterior using both `ES` and `SIES`

# %%
X_ES_ert = X.copy()
Y_ES_ert = Y.copy()
smoother_es = ies.ES()
smoother_es.fit(Y_ES_ert, d.sd.values, d.value.values, noise=E)
X_ES_ert = smoother_es.update(X_ES_ert)

X_IES_ert = X.copy()
Y_IES_ert = Y.copy()
smoother_ies = ies.SIES(ensemble_size=ensemble_size)
n_ies_iter = 7
for i in range(n_ies_iter):
    smoother_ies.fit(Y_IES_ert, d.sd.values, d.value.values, noise=E)
    X_IES_ert = smoother_ies.update(X_IES_ert)

    _coeff_a = X_IES_ert[0, :]
    _coeff_b = X_IES_ert[1, :]
    _coeff_c = X_IES_ert[2, :]

    _fwd_runs = p_map(
        poly,
        _coeff_a,
        _coeff_b,
        _coeff_c,
        [np.arange(max(x_observations) + 1)] * ensemble_size,
        desc=f"SIES ert iteration {i}",
    )

    Y_IES_ert = np.array(
        [fwd_run[d.index.get_level_values("x").to_list()] for fwd_run in _fwd_runs]
    ).T


# %% [markdown]
# ## Plots to compare results


# %%
def plot_posterior(ax, posterior, method):
    for i, param in enumerate(["a", "b", "c"]):
        ax[i].set_title(param)
        ax[i].hist(posterior[i, :], label=f"{method} posterior", alpha=0.5)
        ax[i].legend()

    fig.tight_layout()
    return ax


fig, ax = plt.subplots(nrows=3)
fig.set_size_inches(8, 8)

for i in range(3):
    ax[i].hist(X[i, :], label="prior")
ax[0].axvline(a_t, color="k", linestyle="--", label="truth")
ax[1].axvline(b_t, color="k", linestyle="--", label="truth")
ax[2].axvline(c_t, color="k", linestyle="--", label="truth")

plot_posterior(ax, X_ES_ert, method="ES ert")
_ = plot_posterior(ax, X_IES_ert, method=f"SIES ert ({n_ies_iter})")

# %%



import io
from math import sqrt

import numpy as np
import pytest
from numpy import testing
from scipy.special import erf

import iterative_ensemble_smoother as ies
from iterative_ensemble_smoother.experimental import (
    ensemble_smoother_update_step_row_scaling,
)


# We fix the random seed in the tests for convenience
@pytest.fixture(autouse=True)
def fix_seed(seed=123456789):
    np.random.seed(seed)


def guassian_to_uniform(min_x, max_x):
    """Maps a standard guassian random variable
        to random variable, uniformly distributied between
        min_x and max_x.
    :param min_x: The lower bound on the returned value.
    :param max_x: The upper bound on the returned value.
    """

    def random_variable(x):
        """maps standard normal random variable x to
        uniform between min_x and max_x.
        """
        y = 0.5 * (1 + erf(x / sqrt(2.0)))
        return y * (max_x - min_x) + min_x

    return random_variable


# The following is an example of history matching with the
# iterative_ensemble_smoother library.

# The setup contains a forward model (a second degree polynomial in this case),
# where the coefficents of the polynomial are the model parameters.

# There are 5 time steps t=0,1,2,3,4 and 3 observations at t=0,2,4.
number_of_observations = 3
number_of_time_steps = 5
observation_times = np.array([0, 2, 4])
number_of_realizations = 10

# Before history matching, these observations are predicted by the forward
# model with the priors.


def forward_model(model_parameters):
    """Our :term:`forward_model` is s_0 * t**2 + s_1 * t + s_2 where s_0,
    s_1,s_2 are the model parameters and t is the time.
    """
    return np.array(
        [
            [
                sum(
                    parameter * time ** (2 - i)
                    for i, parameter in enumerate(model_parameters)
                )
            ]
            for time in range(number_of_time_steps)
        ],
    ).reshape((number_of_time_steps, number_of_realizations))


# The priors at t=0,2,4 are assumed uniform in [0,1], [0,2] and [0,5]
# respectively

priors = [
    guassian_to_uniform(0, 1),
    guassian_to_uniform(0, 2),
    guassian_to_uniform(0, 5),
]

# As input to the history matching we have the following
# observed values. These would normally be historic measurements.
observation_values = np.array([2.8532509308, 7.20311703432, 21.3864899107])
# The observed values have the following measurement errors
observation_errors = np.array([0.5 * (x + 1) for x, _ in enumerate(observation_values)])


@pytest.fixture
def initial_A():
    """Initial guess at parameters for the ensemble"""
    return np.random.normal(0, 1, size=(3, number_of_realizations))


@pytest.fixture
def initial_responses(initial_A):
    """The initial responses from the model"""
    return forward_model([prior(x) for prior, x in zip(priors, initial_A)])


@pytest.fixture
def initial_S(initial_responses):
    """S is the matrix of responses for each observation"""
    return initial_responses[observation_times]


def to_csv(nparray):
    buffer = io.StringIO()
    np.savetxt(
        buffer, np.round(nparray, 9), delimiter=",", newline="\n", encoding="utf8"
    )
    return buffer.getvalue()


def test_iterative_ensemble_smoother_update_step(snapshot, initial_A, initial_S):
    # performing an update step gives us a new A matrix with updated parameters
    # for the ensemble
    rng = np.random.default_rng(12345)
    noise = rng.standard_normal(size=(len(observation_values), initial_A.shape[1]))
    smoother = ies.SIES(ensemble_size=initial_A.shape[1])
    smoother.fit(
        initial_S,
        observation_errors,
        observation_values,
        noise=noise,
        step_length=1.0,
        param_ensemble=initial_A,
    )
    new_A = smoother.update(initial_A)
    assert new_A.shape == initial_A.shape
    assert new_A.dtype == initial_A.dtype
    snapshot.assert_match(
        to_csv(new_A), "test_iterative_ensemble_smoother_update_step.csv"
    )


class RowScaling:
    def multiply(self, A, _):
        return A


def test_ensemble_smoother_update_step_with_rowscaling(snapshot, initial_A, initial_S):
    ((new_A, _),) = ensemble_smoother_update_step_row_scaling(
        initial_S,
        [(initial_A, RowScaling())],
        observation_errors,
        observation_values,
    )
    assert new_A.shape == initial_A.shape
    assert new_A.dtype == initial_A.dtype
    snapshot.assert_match(to_csv(new_A), "test_ensemble_smoother_update_step.csv")


def test_ensemble_smoother_update_step(snapshot, initial_A, initial_S):
    # performing an update step gives us a new A matrix with updated parameters
    # for the ensemble
    rng = np.random.default_rng(12345)
    noise = rng.standard_normal(size=(len(observation_values), initial_A.shape[1]))
    ensemble_size = initial_A.shape[1]
    smoother = ies.SIES(ensemble_size)
    smoother.fit(
        initial_S,
        observation_errors,
        observation_values,
        step_length=1.0,
        noise=noise,
        param_ensemble=initial_A,
    )
    new_A = smoother.update(initial_A)
    assert new_A.shape == initial_A.shape
    assert new_A.dtype == initial_A.dtype
    snapshot.assert_match(to_csv(new_A), "test_ensemble_smoother_update_step.csv")


def test_get_steplength():
    expected = [
        7.762203155904597862e-01,
        5.999999999999999778e-01,
        4.889881574842309675e-01,
        4.190550788976149521e-01,
        3.750000000000000000e-01,
        3.472470393710577197e-01,
        3.297637697244037436e-01,
        3.187499999999999778e-01,
        3.118117598427644355e-01,
        3.074409424311009276e-01,
    ]
    iterative_es = ies.SIES(0)
    steplengths = [iterative_es._get_steplength(i) for i in range(10)]
    testing.assert_array_equal(expected, steplengths)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Example: Estimating parameters of an anharmonic oscillator
#
# The anharnomic oscillator can be modelled by a non-linear partial differential
# equation as described in section 6.4.3 of the book Fundamentals of Algorithms
# and Data Assimilation by Mark Asch, Marc Bocquet and Maëlle Nodet.

# %%
# Simple plotting of forward-model with a single response and parameters
from matplotlib import pyplot as plt


def plot_result(A, response_x_axis, trans_func=lambda x: x, priors=[]):
    responses = forward_model(A, priors, response_x_axis)
    plt.rcParams["figure.figsize"] = [15, 4]
    _, axs = plt.subplots(1, 1 + len(A))

    axs[0].plot(response_x_axis, responses)
    for i, param in enumerate(A):
        A_trans = np.array([trans_func(v, *priors[i]) for v in param])
        axs[i + 1].hist(A_trans, bins=10)
    plt.show()


# %% [markdown]
# ## Setup

# %%
# Oscilator example
import numpy as np
from math import sqrt
from scipy.special import erf


def _generate_observations(K):
    x = _evaluate(omega=3.5e-2, lmbda=3e-4, K=K)
    rng = np.random.default_rng(12345)
    nobs = 50
    obs_points = np.linspace(0, K, nobs, endpoint=False, dtype=int)

    obs_with_std = np.zeros(shape=(len(obs_points), 2))

    for obs_idx, obs_point in enumerate(obs_points):
        # Set observation error's standard deviation to some
        # percentage of the amplitude of x with a minimum of, e.g., 1.
        obs_std = max(1, 0.02 * abs(x[obs_point]))
        obs_with_std[obs_idx, 0] = x[obs_point] + rng.normal(loc=0.0, scale=obs_std)
        obs_with_std[obs_idx, 1] = obs_std
    return obs_with_std, obs_points


def _evaluate(omega, lmbda, K):
    x = np.zeros(K)
    x[0] = 0
    x[1] = 1

    # Looping from 2 because we have initial conditions at k=0 and k=1.
    for k in range(2, K - 1):
        M = np.array([[2 + omega**2 - lmbda**2 * x[k] ** 2, -1], [1, 0]])
        u = np.array([x[k], x[k - 1]])
        u = M @ u
        x[k + 1] = u[0]
        x[k] = u[1]

    return x


def uniform(x, min_x, max_x):
    y = 0.5 * (1 + erf(x / sqrt(2.0)))
    return y * (max_x - min_x) + min_x


def forward_model(A, prior, response_x_axis):
    responses = []
    for [omega, lmbda] in A.T:
        r = _evaluate(
            omega=uniform(omega, *prior[0]),
            lmbda=uniform(lmbda, *prior[1]),
            K=len(response_x_axis),
        )
        responses.append(r)
    return np.array(responses).T


response_x_axis = range(2500)
realizations = 100

observations, observation_x_axis = _generate_observations(len(response_x_axis))
observation_values = observations[:, 0]
observation_errors = observations[:, 1]

A = np.random.normal(0, 1, size=(2, realizations))

priors = [(2.5e-2, 4.5e-2), (2.0e-4, 4.0e-4)]
plot_result(A, response_x_axis, uniform, priors)


# %% [markdown]
# ## Update step

# %%
import numpy as np
import iterative_ensemble_smoother as ies

plot_result(A, response_x_axis, uniform, priors)

responses_before = forward_model(A, priors, response_x_axis)
Y = responses_before[observation_x_axis]

smoother = ies.ES()
smoother.fit(Y, observation_errors, observation_values)
new_A = smoother.update(A)

plot_result(new_A, response_x_axis, uniform, priors)

# %% [markdown]
# ## Iterative smoother

# %%
import numpy as np
from matplotlib import pyplot as plt
import iterative_ensemble_smoother as ies


def iterative_smoother():
    A_current = np.copy(A)
    iterations = 4
    smoother = ies.SIES(realizations)

    for _ in range(iterations):
        plot_result(A_current, response_x_axis, uniform, priors)

        responses_before = forward_model(A_current, priors, response_x_axis)
        Y = responses_before[observation_x_axis]

        smoother.fit(Y, observation_errors, observation_values)
        A_current = smoother.update(A_current)
    plot_result(A_current, response_x_axis, uniform, priors)


iterative_smoother()

# %% [markdown]
# ## ES-MDA (Multiple Data Assimilation - Ensemble Smoother)

# %%
import numpy as np
from matplotlib import pyplot as plt
import iterative_ensemble_smoother as ies


def es_mda():
    A_current = np.copy(A)
    weights = [8, 4, 2, 1]
    length = sum(1.0 / x for x in weights)

    smoother = ies.ES()
    for weight in weights:
        plot_result(A_current, response_x_axis, uniform, priors)

        responses_before = forward_model(A_current, priors, response_x_axis)
        Y = responses_before[observation_x_axis]

        observation_errors_scaled = observation_errors * sqrt(weight * length)
        smoother.fit(Y, observation_errors_scaled, observation_values)
        A_current = smoother.update(A_current)
    plot_result(A_current, response_x_axis, uniform, priors)


es_mda()

# %%



""" Implementation of the iterative ensemble smoother history matching algorithms
from Evensen et al. "Efficient Implementation of an Iterative Ensemble Smoother
for Data Assimilation and Reservoir History Matching"
https://www.frontiersin.org/articles/10.3389/fams.2019.00047/full
"""

from ._ies import InversionType
from ._iterative_ensemble_smoother import (
    ES,
    SIES,
)

__all__ = [
    "ES",
    "SIES",
    "InversionType",
]



from subprocess import check_output

project = "iterative_ensemble_smoother"
copyright = "2022, Equinor"
author = "Equinor"
release = "0.0.1"


check_output(["jupytext", "Polynomial.py", "-o", "Polynomial.ipynb"])

check_output(["jupytext", "Oscillator.py", "-o", "Oscillator.ipynb"])

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "nbsphinx",
    "sphinxcontrib.bibtex",
]
bibtex_bibfiles = ["refs.bib"]
language = "python"
html_theme = "sphinx_rtd_theme"



from __future__ import annotations
from typing import Optional, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import numpy.typing as npt

rng = np.random.default_rng()

from ._ies import InversionType, make_D, make_E, create_coefficient_matrix
from iterative_ensemble_smoother.utils import (
    _validate_inputs,
    _create_errors,
)


def _response_projection(
    param_ensemble: npt.NDArray[np.double],
) -> npt.NDArray[np.double]:
    """A^+A projection is necessary when the parameter matrix has fewer rows than
    columns, and when the forward model is non-linear. Section 2.4.3
    """
    ensemble_size = param_ensemble.shape[1]
    A = (param_ensemble - param_ensemble.mean(axis=1, keepdims=True)) / np.sqrt(
        ensemble_size - 1
    )
    projection: npt.NDArray[np.double] = np.linalg.pinv(A) @ A
    return projection


class SIES:
    """SIES performs the update step of the Subspace Iterative Ensemble Smoother
    algorithm. See `Evensen[1]`_.

    :param ensemble_size: The number of realizations in the ensemble model.
    :param max_steplength: parameter used to tweaking the step length.
    :param min_steplength: parameter used to tweaking the step length.
    :param dec_steplength: parameter used to tweaking the step length.
    """

    def __init__(
        self,
        ensemble_size: int,
        *,
        max_steplength: float = 0.6,
        min_steplength: float = 0.3,
        dec_steplength: float = 2.5,
    ):
        self._initial_ensemble_size = ensemble_size
        self.iteration_nr = 1
        self.max_steplength = max_steplength
        self.min_steplength = min_steplength
        self.dec_steplength = dec_steplength
        self.coefficient_matrix = np.zeros(shape=(ensemble_size, ensemble_size))

    def _get_steplength(self, iteration_nr: int) -> float:
        """
        This is an implementation of Eq. (49), which calculates a suitable step length for
        the update step, from the book:

        Geir Evensen, Formulating the history matching problem with consistent error statistics,
        Computational Geosciences (2021) 25:945 –970
        """
        steplength = self.min_steplength + (
            self.max_steplength - self.min_steplength
        ) * pow(2, -(iteration_nr - 1) / (self.dec_steplength - 1))
        return steplength

    def fit(
        self,
        response_ensemble: npt.NDArray[np.double],
        observation_errors: npt.NDArray[np.double],
        observation_values: npt.NDArray[np.double],
        *,
        noise: Optional[npt.NDArray[np.double]] = None,
        truncation: float = 0.98,
        step_length: Optional[float] = None,
        ensemble_mask: Optional[npt.NDArray[np.bool_]] = None,
        inversion: InversionType = InversionType.EXACT,
        param_ensemble: Optional[npt.NDArray[np.double]] = None,
    ) -> None:
        """Perform one step of the iterative ensemble smoother algorithm

        :param response_ensemble: Matrix of responses from the :term:`forward model`.
            Has shape (number of observations, number of realizations).
            (Y in Evensen et. al)
        :param observation_errors: 1D array of measurement errors (standard deviations)
                                   for each observation, or covariance matrix
                                   if errors are correlated.
        :param observation_values: 1D array of observations.
        :param noise: Optional noise matrix with the same shape as response matrix.
        :param truncation: float used to determine the number of significant singular
            values. Defaults to 0.98 (ie. 98% significant values).
        :param step_length: The step length to be used in the algorithm,
            defaults to using the method described in Eq. 49 Geir Evensen,
            Formulating the history matching problem with consistent error
            statistics, Computational Geosciences (2021) 25:945 –970
        :param ensemble_mask: 1D array describing which realizations are active. Defaults
            to all active. Inactive realizations are ignored.
        :param inversion: The type of subspace inversion used in the algorithm, defaults
            to exact.
        :param param_ensemble: All parameters input to dynamical model used to
            generate responses.
            Must be passed if total number of parameters is
            less than ensemble_size - 1 and the dynamical model is non-linear.
        """

        R, observation_errors = _create_errors(observation_errors, inversion)

        _validate_inputs(
            response_ensemble,
            noise,
            observation_errors,
            observation_values,
            param_ensemble=param_ensemble,
        )

        num_obs = len(observation_values)
        ensemble_size = response_ensemble.shape[1]
        if step_length is None:
            step_length = self._get_steplength(self.iteration_nr)
        if noise is None:
            noise = rng.standard_normal(size=(num_obs, ensemble_size))

        E = make_E(observation_errors, noise)
        D = make_D(observation_values, E, response_ensemble)
        D = (D.T / observation_errors).T
        E = (E.T / observation_errors).T

        if param_ensemble is not None:
            projected_response = response_ensemble @ _response_projection(
                param_ensemble
            )
            response_ensemble = (projected_response.T / observation_errors).T
        else:
            response_ensemble = (response_ensemble.T / observation_errors).T

        if ensemble_mask is None:
            ensemble_mask = np.array([True] * ensemble_size)

        self.ensemble_mask = ensemble_mask

        W: npt.NDArray[np.double] = create_coefficient_matrix(
            (response_ensemble - response_ensemble.mean(axis=1, keepdims=True))
            / np.sqrt(ensemble_size - 1),
            R,
            E,
            D,
            inversion,
            truncation,
            self.coefficient_matrix[ensemble_mask, :][:, ensemble_mask],
            step_length,
        )

        if np.isnan(W).sum() != 0:
            raise ValueError(
                "Fit produces NaNs. Check your response matrix for outliers or use an inversion type with truncation."
            )

        self.iteration_nr += 1

        self.coefficient_matrix[np.outer(ensemble_mask, ensemble_mask)] = W.ravel()

    def update(self, param_ensemble: npt.NDArray[np.double]) -> npt.NDArray[np.double]:
        # Line 9 of Algorithm 1
        ensemble_size = self.ensemble_mask.sum()
        I = np.identity(ensemble_size)
        W = self.coefficient_matrix[self.ensemble_mask, :][:, self.ensemble_mask]
        transition_matrix: npt.NDArray[np.double] = I + W / np.sqrt(ensemble_size - 1)
        return param_ensemble @ transition_matrix

    def __repr__(self) -> str:
        return (
            f"SIES(ensemble_size={self._initial_ensemble_size}, "
            f"max_steplength={self.max_steplength}, min_steplength={self.min_steplength}, "
            f"dec_steplength={self.dec_steplength})"
        )


class ES:
    def __init__(self) -> None:
        self.smoother: Optional[SIES] = None

    def fit(
        self,
        response_ensemble: npt.NDArray[np.double],
        observation_errors: npt.NDArray[np.double],
        observation_values: npt.NDArray[np.double],
        *,
        noise: Optional[npt.NDArray[np.double]] = None,
        truncation: float = 0.98,
        inversion: InversionType = InversionType.EXACT,
        param_ensemble: Optional[npt.NDArray[np.double]] = None,
    ) -> None:
        self.smoother = SIES(ensemble_size=response_ensemble.shape[1])
        self.smoother.fit(
            response_ensemble,
            observation_errors,
            observation_values,
            noise=noise,
            truncation=truncation,
            step_length=1.0,
            ensemble_mask=None,
            inversion=inversion,
            param_ensemble=param_ensemble,
        )

    def update(self, param_ensemble: npt.NDArray[np.double]) -> npt.NDArray[np.double]:
        assert self.smoother is not None
        return self.smoother.update(param_ensemble)

    def __repr__(self) -> str:
        return "ES()"



"""
Contains (publicly available, but not officially supported) experimental
features of iterative_ensemble_smoother
"""
import numpy as np

rng = np.random.default_rng()

from iterative_ensemble_smoother.utils import (
    _create_errors,
)

from ._ies import InversionType, make_D, make_E, create_coefficient_matrix


def ensemble_smoother_update_step_row_scaling(
    response_ensemble,
    A_with_row_scaling,
    observation_errors,
    observation_values,
    noise=None,
    truncation=0.98,
    inversion=InversionType.EXACT,
):
    """This is an experimental feature."""
    ensemble_size = response_ensemble.shape[1]
    if noise is None:
        num_obs = len(observation_values)
        noise = rng.standard_normal(size=(num_obs, ensemble_size))

    R, observation_errors = _create_errors(observation_errors, inversion)

    E = make_E(observation_errors, noise)
    D = make_D(observation_values, E, response_ensemble)
    D = (D.T / observation_errors).T
    E = (E.T / observation_errors).T
    response_ensemble = (response_ensemble.T / observation_errors).T
    for A, row_scale in A_with_row_scaling:
        W = create_coefficient_matrix(
            (response_ensemble - response_ensemble.mean(axis=1, keepdims=True))
            / np.sqrt(ensemble_size - 1),
            R,
            E,
            D,
            inversion,
            truncation,
            np.zeros((ensemble_size, ensemble_size)),
            1.0,
        )
        I = np.identity(ensemble_size)
        transition_matrix = I + W / np.sqrt(ensemble_size - 1)
        row_scale.multiply(A, transition_matrix)
    return A_with_row_scaling



from iterative_ensemble_smoother import ES, SIES, InversionType
import numpy as np
import pytest
import re


def test_that_repr_can_be_created():
    smoother = SIES(100)
    smoother_repr = eval(repr(smoother))
    assert isinstance(smoother_repr, SIES)

    smoother = ES()
    smoother_repr = eval(repr(smoother))
    assert isinstance(smoother_repr, ES)


def test_that_bad_inputs_cause_nice_error_messages():
    ensemble_size = 10
    num_obs = 4
    num_params = 3
    param_ensemble = np.random.normal(size=(num_obs, ensemble_size))
    Y = np.ones(shape=(num_obs, ensemble_size))
    obs_errors = np.ones(num_obs)
    obs_values = np.ones(num_obs)
    noise = np.zeros_like(Y)

    with pytest.raises(
        ValueError,
        match="noise and response_ensemble must have the same number of columns",
    ):
        _ = SIES(ensemble_size).fit(Y, obs_errors, obs_values, noise=noise[:, 1:])

    with pytest.raises(
        ValueError,
        match="noise and response_ensemble must have the same number of rows",
    ):
        _ = SIES(ensemble_size).fit(Y, obs_errors, obs_values, noise=noise[1:, :])

    with pytest.raises(
        ValueError,
        match=re.escape(
            "response_ensemble must be a matrix of size (number of responses by number of realizations)"
        ),
    ):
        _ = SIES(ensemble_size).fit(Y.ravel(), obs_errors, obs_values, noise=noise)

    with pytest.raises(
        ValueError,
        match=re.escape(
            "observation_errors and observation_values must have the same number of elements"
        ),
    ):
        _ = SIES(ensemble_size).fit(Y, obs_errors[1:], obs_values, noise=noise)

    with pytest.raises(
        ValueError,
        match=re.escape(
            "observation_values must have the same number of elements as there are responses"
        ),
    ):
        _ = SIES(ensemble_size).fit(Y, obs_errors[1:], obs_values[1:], noise=noise)

    with pytest.raises(
        ValueError,
        match="param_ensemble and response_ensemble must have the same number of columns",
    ):
        _ = SIES(ensemble_size).fit(
            Y,
            obs_errors,
            obs_values,
            noise=noise,
            param_ensemble=param_ensemble[:, : ensemble_size - 2],
        )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "parameter_ensemble must be a matrix of size (number of parameters by number of realizations)"
        ),
    ):
        _ = SIES(ensemble_size).fit(
            Y,
            obs_errors,
            obs_values,
            noise=noise,
            param_ensemble=param_ensemble[0, :].ravel(),
        )


def test_that_nans_produced_due_to_outliers_in_responses_are_handled():
    rng = np.random.default_rng()
    # Creating response matrix with large outlier that will
    # lead to NaNs.
    response_ensemble = np.array([[1, 1, 1e19], [1, 10, 100]])
    noise = rng.normal(size=(2, 3))
    obs_error = np.array([1, 2])
    obs_value = np.array([10, 20])
    smoother = ES()

    with pytest.raises(
        ValueError,
        match="Fit produces NaNs. Check your response matrix for outliers or use an inversion type with truncation.",
    ):
        smoother.fit(response_ensemble, obs_error, obs_value, noise=noise)

    # Running with an inversion type that does truncation does not produce NaNs.
    param_ensemble = np.array([[1, 2, 3]])
    smoother.fit(
        response_ensemble,
        obs_error,
        obs_value,
        noise=noise,
        inversion=InversionType.EXACT_R,
    )
    param_ensemble = smoother.update(param_ensemble)
    assert np.isnan(param_ensemble).sum() == 0



from glob import glob
from os import path
from pathlib import Path
from subprocess import check_output

from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup

check_output(["conan", "install", "."])

ext_modules = [
    Pybind11Extension(
        "iterative_ensemble_smoother._ies",
        sorted(glob("src/iterative_ensemble_smoother/*.cpp")),
        cxx_std=17,
        include_dirs=[
            path.join(path.dirname(__file__), "src/iterative_ensemble_smoother/"),
        ],
        extra_compile_args=Path("conanbuildinfo.args").read_text().split(),
    ),
]

setup(ext_modules=ext_modules)
